from typing import Iterable, Union, Tuple, List
import math
import numpy as np
import operator

import poisson_disc

from ._imageloaders import ImageCollection, ImageBase


class PatchSampler(object):
    """Patch sampling algorithm to retrieve image patches/windows from images.

    If the image used for extracting patches has a mask associated to it, only
    patches from masked regions of the image are retrieved.

    Parameters
    ----------
    patch_size : Union[int, Iterable[int], dict]
        Size in pixels of the patches extracted on each axis. Only rectangular
        patches (hyper-cuboids) are supported by now. If a single int is
        passed, that size is used for all dimensions. If an iterable (list,
        tuple) is passed, each value will be assigned to the corresponding axes
         in `spatial_axes`, the size of `patch_size` must match the lenght of
        `spatial_axes'. If a dict is passed, this should have at least the size
         of the patch of the axes listed in `spatial_axes`.  Use the same
        convention as how Zarr structure array chunks in order to handle patch
        shapes and channels correctly.
    stride : Union[int, Iterable[int], dict, None]
        Distance in pixels of the movement of the sampling sliding window.
        If `stride` is less than `patch_size` for an axis, patches will have an
        overlap between them. This is usuful in inference mode for avoiding
        edge artifacts. If None is passed, the `patch_size` will be used as
        `stride`.
    pad : Union[int, Iterable[int], dict, None]
        Padding in pixels added to the extracted patch at each specificed axis.
    min_area : Union[int, float]
        Minimum patch area covered by the mask to consider it samplable. A
        number in range [0, 1) will be used as percentage of the patch size. A
        number >= 1 will be considered as the number of minimum patch pixels
        covered by the mask.
    spatial_axes : str
        The spatial axes from where patches can be extracted.
    allow_incomplete_patches : bool
        Allow to retrieve patches that are smaller than the patch size. This is
        the case of samples at the edge of the image that are usually smaller
        than the specified patch size.
    """
    def __init__(self, patch_size: Union[int, Iterable[int], dict],
                 stride: Union[int, Iterable[int], dict, None] = None,
                 pad: Union[int, Iterable[int], dict, None] = None,
                 min_area: Union[int, float] = 1,
                 spatial_axes: str = "ZYX",
                 allow_incomplete_patches: bool = False):
        # The maximum chunk sizes are used to generate a reference sampling
        # position array used fo every sampled chunk.
        self._max_chunk_size = {ax: 0 for ax in spatial_axes}

        if isinstance(patch_size, (list, tuple)):
            if len(patch_size) != len(spatial_axes):
                raise ValueError(f"The size of `patch_size` must match the "
                                 f"number of axes in `spatial_axes`, got "
                                 f"{len(patch_size)} for {spatial_axes}")

            patch_size = {ax: ps for ax, ps in zip(spatial_axes, patch_size)}

        elif isinstance(patch_size, int):
            patch_size = {ax: patch_size for ax in spatial_axes}

        elif not isinstance(patch_size, dict):
            raise ValueError(f"Patch size must be a dictionary specifying the"
                             f" patch size of each axes, an iterable (list, "
                             f"tuple) with the same order as the spatial axes,"
                             f" or an integer for a cubic patch. Received "
                             f"{patch_size} of type {type(patch_size)}")

        if isinstance(stride, (list, tuple)):
            if len(stride) != len(spatial_axes):
                raise ValueError(f"The size of `stride` must match the "
                                 f"number of axes in `spatial_axes`, got "
                                 f"{len(stride)} for {spatial_axes}")

            stride = {ax: st for ax, st in zip(spatial_axes, stride)}

        elif isinstance(stride, int):
            stride = {ax: stride for ax in spatial_axes}

        elif stride is None:
            stride = patch_size

        elif not isinstance(stride, dict):
            raise ValueError(f"Stride size must be a dictionary specifying the"
                             f" stride step size of each axes, an iterable ("
                             f"list, tuple) with the same order as the spatial"
                             f" axes, or an integer for a cubic patch. "
                             f"Received {stride} of type {type(stride)}")

        if pad is None:
            pad = 0

        if isinstance(pad, (list, tuple)):
            if len(pad) != len(spatial_axes):
                raise ValueError(f"The size of `pad` must match the "
                                 f"number of axes in `spatial_axes`, got "
                                 f"{len(pad)} for {spatial_axes}")

            pad = {ax: st for ax, st in zip(spatial_axes, pad)}

        elif isinstance(pad, int):
            pad = {ax: pad for ax in spatial_axes}

        elif not isinstance(pad, dict):
            raise ValueError(f"Pad size must be a dictionary specifying the"
                             f" numer of pixels added to each axes, an "
                             f"iterable (list, tuple) with the same order as "
                             f"the spatial axes, or an integer for a cubic "
                             f"patch. Received {pad} of type {type(pad)}")

        self.spatial_axes = spatial_axes

        self._patch_size = {ax: patch_size.get(ax, 1) for ax in spatial_axes}
        self._stride = {ax: stride.get(ax, 1) for ax in spatial_axes}
        self._pad = {ax: pad.get(ax, 0) for ax in spatial_axes}

        self._min_area = min_area
        self._allow_incomplete_patches = allow_incomplete_patches

    def _compute_corners(self, coordinates: np.ndarray, scale: np.ndarray
                         ) -> np.ndarray:
        toplefts_corners = []

        dim = coordinates.shape[-1]
        factors = 2 ** np.arange(dim + 1)
        for d in range(2 ** dim):
            corner_value = np.array((d % factors[1:]) // factors[:-1],
                                    dtype=np.float32)
            toplefts_corners.append(
                coordinates + scale * (1 - 1e-4) * corner_value
            )

        corners = np.stack(toplefts_corners)

        return corners

    def _compute_reference_indices(self, reference_coordinates: np.ndarray,
                                   reference_axes_sizes: np.ndarray
                                   ) -> Tuple[List[np.ndarray],
                                              List[Tuple[int]]]:
        reference_per_axis = list(map(
            lambda coords, axis_size: np.concatenate((
                np.full((1, ), fill_value=-float("inf")),
                np.unique(coords),
                np.full((1, ), fill_value=np.max(coords) + axis_size))),
            reference_coordinates.T,
            reference_axes_sizes
        ))

        reference_idx = map(
            lambda coord_axis, ref_axis:
            np.max(np.arange(ref_axis.size)
                   * (coord_axis.reshape(-1, 1) >= ref_axis[None, ...]),
                   axis=1),
            reference_coordinates.T,
            reference_per_axis
        )
        reference_idx = np.stack(tuple(reference_idx), axis=-1)
        reference_idx = reference_idx.reshape(reference_coordinates.T.shape)

        reference_idx = [
            tuple(tls_coord)
            for tls_coord in reference_idx.reshape(-1, len(reference_per_axis))
        ]

        return reference_per_axis, reference_idx

    def _compute_overlap(self, corners_coordinates: np.ndarray,
                         reference_per_axis: np.ndarray) -> Tuple[np.ndarray,
                                                                  np.ndarray]:
        tls_idx = map(
            lambda coord_axis, ref_axis:
            np.max(np.arange(ref_axis.size)
                   * (coord_axis.reshape(-1, 1) >= ref_axis[None, ...]),
                   axis=1),
            np.moveaxis(corners_coordinates, -1, 0),
            reference_per_axis
        )
        tls_idx = np.stack(tuple(tls_idx), axis=-1)
        tls_idx = tls_idx.reshape(corners_coordinates.shape)

        tls_coordinates = map(
            lambda tls_coord, ref_axis: ref_axis[tls_coord],
            np.moveaxis(tls_idx, -1, 0),
            reference_per_axis
        )
        tls_coordinates = np.stack(tuple(tls_coordinates), axis=-1)

        corners_cut = np.maximum(corners_coordinates[0], tls_coordinates[-1])

        dist2cut = np.fabs(corners_coordinates - corners_cut[None])
        coverage = np.prod(dist2cut, axis=-1)

        return coverage, tls_idx

    def _compute_grid(self, chunk_tlbr: dict, mask: ImageBase,
                      patch_size: dict,
                      image_size: dict,
                      min_area: float,
                      allow_incomplete_patches: bool = False):
        mask_scale = np.array([mask.scale.get(ax, 1)
                               for ax in self.spatial_axes],
                              dtype=np.float32)

        image_scale = np.array([image_size.get(ax, 1) / patch_size.get(ax, 1)
                                for ax in self.spatial_axes],
                               dtype=np.float32)

        round_fn = math.ceil if allow_incomplete_patches else math.floor

        image_blocks = [
            round_fn(
                (
                    min(image_size.get(ax, 1),
                        chunk_tlbr[ax].stop
                        if chunk_tlbr[ax].stop is not None
                        else float("inf"))
                    - (chunk_tlbr[ax].start
                       if chunk_tlbr[ax].start is not None
                       else 0)
                ) / patch_size.get(ax, 1))
            for ax in self.spatial_axes
        ]

        if min(image_blocks) == 0:
            return []

        image_scale = np.array([patch_size.get(ax, 1)
                                for ax in self.spatial_axes],
                               dtype=np.float32)
        image_coordinates = np.array(list(np.ndindex(*image_blocks)),
                                     dtype=np.float32)

        image_coordinates *= image_scale

        mask_scale = 1 / np.array([mask.scale.get(ax, 1)
                                   for ax in self.spatial_axes],
                                  dtype=np.float32)

        mask_coordinates = list(np.nonzero(mask[:]))
        for ax_i, ax in enumerate(self.spatial_axes):
            if ax not in mask.axes:
                mask_coordinates.insert(
                    ax_i,
                    np.zeros(mask_coordinates[0].size)
                )

        mask_coordinates = np.stack(mask_coordinates, dtype=np.float32).T
        mask_coordinates *= mask_scale[None, ...]

        # Filter out mask coordinates outside the current selected chunk
        chunk_tl_coordinates = np.array(
            [chunk_tlbr[ax].start if chunk_tlbr[ax].start is not None else 0
             for ax in self.spatial_axes],
            dtype=np.float32
        )

        chunk_br_coordinates = np.array(
            [chunk_tlbr[ax].stop
             if chunk_tlbr[ax].stop is not None
             else float('inf')
             for ax in self.spatial_axes],
            dtype=np.float32
        )

        in_chunk = np.all(
            np.bitwise_and(
                mask_coordinates > (chunk_tl_coordinates - mask_scale - 1e-4),
                mask_coordinates < chunk_br_coordinates + 1e-4
            ),
            axis=1
        )
        mask_coordinates = mask_coordinates[in_chunk]

        # Translate the mask coordinates to the origin for comparison with
        # image coordinates.
        mask_coordinates -= chunk_tl_coordinates

        if all(map(operator.ge, image_scale, mask_scale)):
            mask_corners = self._compute_corners(mask_coordinates, mask_scale)

            (reference_per_axis,
             reference_idx) =\
                self._compute_reference_indices(image_coordinates, image_scale)

            (coverage,
             corners_idx) = self._compute_overlap(mask_corners,
                                                  reference_per_axis)

            covered_indices = [
                reference_idx.index(tuple(idx))
                if tuple(idx) in reference_idx else len(reference_idx)
                for idx in corners_idx.reshape(-1, len(self.spatial_axes))
            ]

            patches_coverage = np.bincount(covered_indices,
                                           weights=coverage.flatten(),
                                           minlength=len(reference_idx) + 1)
            patches_coverage = patches_coverage[:-1]

        else:
            image_corners = self._compute_corners(image_coordinates,
                                                  image_scale)

            (reference_per_axis,
             reference_idx) = self._compute_reference_indices(mask_coordinates,
                                                              mask_scale)

            (coverage,
             corners_idx) = self._compute_overlap(image_corners,
                                                  reference_per_axis)

            covered_indices = np.array([
                tuple(idx) in reference_idx
                for idx in corners_idx.reshape(-1, len(self.spatial_axes))
            ]).reshape(coverage.shape)

            patches_coverage = np.sum(covered_indices * coverage, axis=0)

        minimum_covered_tls = image_coordinates[patches_coverage > min_area]
        minimum_covered_tls = minimum_covered_tls.astype(np.int64)

        return minimum_covered_tls

    def _compute_valid_toplefts(self, chunk_tlbr: dict, mask: ImageBase,
                                patch_size: dict,
                                **kwargs):
        return self._compute_grid(chunk_tlbr, mask, patch_size, **kwargs)

    def _compute_toplefts_slices(self, chunk_tlbr: dict,
                                 valid_mask_toplefts: np.ndarray,
                                 patch_size: dict,
                                 pad: Union[dict, None] = None):
        if pad is None:
            pad = {ax: 0 for ax in self.spatial_axes}

        toplefts = [
            {ax: slice(
                (chunk_tlbr[ax].start if chunk_tlbr[ax].start is not None
                 else 0) + tls[self.spatial_axes.index(ax)]
                - pad[ax],
                (chunk_tlbr[ax].start if chunk_tlbr[ax].start is not None
                 else 0) + tls[self.spatial_axes.index(ax)] + patch_size[ax]
                + pad[ax])
             for ax in self.spatial_axes
             }
            for tls in valid_mask_toplefts
        ]

        return toplefts

    def compute_chunks(self,
                       image_collection: ImageCollection) -> Iterable[dict]:
        """Compute the top-left positions of the chunks in which the image is
        divided before taking patches from them.

        The chunk size should match the zarr's chunk sizes to make the patch
        extraction more efficient.

        Parameters
        ----------
        image : zarrdataset.ImageCollection
            The image from where the patches are extracted.

        Returns
        -------
        chunks_grids : list of tuples of slices
            Each valid chunk is returned in form of a slice that can be
            extracted from the image. The slices are stored as tuples, with one
            slice per axis.
        """
        # Use the mask as base to determine the valid patches that can be
        # retrieved from the image.
        image = image_collection.collection[image_collection.reference_mode]
        mask = image_collection.collection[image_collection.mask_mode]

        # This computes a chunk size in terms of the patch size instead of the
        # original array chunk size.
        spatial_chunk_sizes = {
            ax: (self._stride[ax]
                 * max(1, math.ceil(chk / self._stride[ax])))
            for ax, chk in zip(image.axes, image.chunk_size)
            if ax in self.spatial_axes
        }

        image_size = {ax: s for ax, s in zip(image.axes, image.shape)}

        self._max_chunk_size = {
            ax: (min(max(self._max_chunk_size[ax],
                         spatial_chunk_sizes[ax]),
                     image_size[ax]))
            if ax in image.axes else 1
            for ax in self.spatial_axes
        }

        chunk_tlbr = {ax: slice(None) for ax in self.spatial_axes}

        valid_mask_toplefts = self._compute_grid(
            chunk_tlbr,
            mask,
            self._max_chunk_size,
            image_size,
            min_area=1,
            allow_incomplete_patches=True
        )

        chunks_slices = self._compute_toplefts_slices(
            chunk_tlbr,
            valid_mask_toplefts=valid_mask_toplefts,
            patch_size=self._max_chunk_size
        )

        return chunks_slices

    def compute_patches(self, image_collection: ImageCollection,
                        chunk_tlbr: dict) -> Iterable[dict]:
        image = image_collection.collection[image_collection.reference_mode]
        mask = image_collection.collection[image_collection.mask_mode]
        image_size = {ax: s for ax, s in zip(image.axes, image.shape)}

        stride = {
            ax: self._stride.get(ax, 1) if image_size.get(ax, 1) > 1 else 1
            for ax in self.spatial_axes
        }

        patch_size = {
            ax: self._patch_size.get(ax, 1) if image_size.get(ax, 1) > 1 else 1
            for ax in self.spatial_axes
        }

        pad = {
            ax: self._pad.get(ax, 0) if image_size.get(ax, 1) > 1 else 0
            for ax in self.spatial_axes
        }

        min_area = self._min_area
        if min_area < 1:
            min_area *= np.prod(list(patch_size.values()))

        valid_mask_toplefts = self._compute_valid_toplefts(
            chunk_tlbr,
            mask,
            stride,
            image_size=image_size,
            min_area=min_area,
            allow_incomplete_patches=self._allow_incomplete_patches
        )

        patches_slices = self._compute_toplefts_slices(
            chunk_tlbr,
            valid_mask_toplefts=valid_mask_toplefts,
            patch_size=patch_size,
            pad=pad
        )

        return patches_slices

    def __repr__(self) -> str:
        """String representation of classes derived from PatchSampler.
        """
        return (f"{type(self)} for sampling patches of size "
                f"{self._patch_size}.")


class BlueNoisePatchSampler(PatchSampler):
    """Patch sampler that retrieves patches from coordinates sampled using the
    Bridson sampling algorithm, also known as Blue-noise sampling algorithm.

    Parameters
    ----------
    patch_size : int, iterable, dict
        Size in pixels of the patches extracted. Only squared patches are
        supported by now.
    resample_positions : bool
        Whether positions should be resampled for each new chunk or not.
    allow_overlap : bool
        Whether overlapping of patches is allowed or not.
    """
    def __init__(self, patch_size: Union[int, Iterable[int], dict],
                 resample_positions=False,
                 allow_overlap=False,
                 **kwargs):
        super(BlueNoisePatchSampler, self).__init__(patch_size, **kwargs)
        self._base_chunk_tls = None
        self._resample_positions = resample_positions
        self._allow_overlap = allow_overlap

    def compute_sampling_positions(self, force=False) -> None:
        """Compute the sampling positions using blue-noise sampling.

        Parameters
        ----------
        force: bool
            Whether force resampling positions, or reuse existing positions.
        """
        if self._base_chunk_tls is not None and not force:
            return

        radius = max(self._patch_size.values())
        if self._allow_overlap:
            radius *= math.sqrt(2)

        samplable_axes = [ax for ax in self.spatial_axes
                          if self._max_chunk_size[ax] > radius]
        dims = [self._max_chunk_size[ax] - radius for ax in samplable_axes]

        if len(samplable_axes):
            base_chunk_tls = poisson_disc.Bridson_sampling(dims=np.array(dims),
                                                           radius=radius,
                                                           k=30)
            base_chunk_tls = base_chunk_tls.astype(np.int64)

            base_chunk_tls = [
                base_chunk_tls[:, samplable_axes.index(ax)]
                if ax in samplable_axes else np.zeros(len(base_chunk_tls),
                                                      dtype=np.int64)
                for ax in self.spatial_axes
            ]
            self._base_chunk_tls = np.stack(base_chunk_tls).T

        else:
            self._base_chunk_tls = np.zeros((1, len(self.spatial_axes)),
                                            dtype=np.int64)

    def _compute_valid_toplefts(self, chunk_tlbr: dict, mask: ImageBase,
                                patch_size: dict,
                                **kwargs):
        self.compute_sampling_positions(force=self._resample_positions)

        # Filter sampling positions that does not contain any mask portion.
        spatial_patch_sizes = np.array([
            patch_size.get(ax, 1) for ax in self.spatial_axes
        ])

        mask_scale = np.array([mask.scale.get(ax, 1)
                               for ax in self.spatial_axes],
                              dtype=np.float32)

        mask_scale = 1 / np.array([mask.scale.get(ax, 1)
                                   for ax in self.spatial_axes],
                                  dtype=np.float32)

        mask_coordinates = list(np.nonzero(mask[:]))
        for ax_i, ax in enumerate(self.spatial_axes):
            if ax not in mask.axes:
                mask_coordinates.insert(
                    ax_i,
                    np.zeros(mask_coordinates[0].size)
                )

        mask_coordinates = np.stack(mask_coordinates, dtype=np.float32).T
        mask_coordinates *= mask_scale[None, ...]

        # Filter out mask coordinates outside the current selected chunk
        chunk_tl_coordinates = np.array(
            [chunk_tlbr[ax].start if chunk_tlbr[ax].start is not None else 0
             for ax in self.spatial_axes],
            dtype=np.float32
        )
        chunk_br_coordinates = np.array(
            [chunk_tlbr[ax].stop
             if chunk_tlbr[ax].stop is not None
             else float('inf')
             for ax in self.spatial_axes],
            dtype=np.float32
        )

        in_chunk = np.all(
            np.bitwise_and(
                mask_coordinates > (chunk_tl_coordinates - mask_scale - 1e-4),
                mask_coordinates < chunk_br_coordinates + 1e-4
            ),
            axis=1
        )
        mask_coordinates = mask_coordinates[in_chunk]

        mask_corners = self._compute_corners(mask_coordinates, mask_scale)

        dist = (mask_corners[None, :, :, :]
                - self._base_chunk_tls[:, None, None, :].astype(np.float32)
                - spatial_patch_sizes[None, None, None, :] / 2)

        mask_samplable_pos, = np.nonzero(
            np.any(
                np.all(dist < spatial_patch_sizes[None, None, None], axis=-1),
                axis=(1, 2)
            )
        )

        toplefts = self._base_chunk_tls[mask_samplable_pos]

        return toplefts



import zarrdataset as zds

from pathlib import Path
import numpy as np
import zarr
from PIL import Image
import tifffile


class ImageTransformTest(zds.MaskGenerator):
    def __init__(self, axes):
        super(ImageTransformTest, self).__init__(axes=axes)

    def _compute_transform(self, image: np.ndarray):
        return image * 2


def input_target_transform(x, t, dtype=np.float32):
    return x.astype(dtype), t.astype(dtype)


def generate_collection(source_axes="CYX", shape=(3, 2048, 2048),
                        chunks=(3, 512, 512),
                        dtype=np.uint8,
                        label=0,
                        mask_threshold=127,
                        data_group="0/0",
                        mask_group="labels/masks/0/0",
                        labels_group="labels/segmentation/0/0",
                        classes_group="labels/classes/0/0",
                        **kwargs):

    img_axes = "CYX"

    rel_axes_order = zds.map_axes_order(source_axes=source_axes,
                                        target_axes=img_axes)

    # Generate a test image on memory
    z1, z2 = np.meshgrid(np.arange(shape[rel_axes_order[-1]],
                                   dtype=np.float32),
                         np.arange(shape[rel_axes_order[-2]],
                                   dtype=np.float32))

    img_channels = []
    for _ in range(shape[source_axes.index("C")]):
        img_channels.append(z1 ** round(np.random.rand() * 3)
                            + z2 ** round(np.random.rand() * 3))

    img = np.stack(img_channels)
    img = img / img.max(axis=(1, 2))[:, None, None]
    img = img * 255
    img = img.astype(dtype)

    # Generate a mask from the image
    mask = img.mean(axis=0) < mask_threshold
    mask = mask[::2, ::2]

    # Label the regions of `img`
    labs = np.log2(img.mean(axis=0) + 1).round().astype(np.uint32)

    new_axes = list(set(source_axes) - set(img_axes))
    permute_order = zds.map_axes_order(source_axes=img_axes,
                                       target_axes=source_axes)
    img = img.transpose(permute_order)
    img = np.expand_dims(img, tuple(source_axes.index(a) for a in new_axes))

    store = zarr.MemoryStore()
    main_grp = zarr.open(store)
    main_grp.create_dataset(data_group, data=img, dtype=dtype,
                            chunks=chunks)

    chunk_height = chunks[rel_axes_order[-2]]
    chunk_width = chunks[rel_axes_order[-1]]

    main_grp.create_dataset(mask_group, data=mask, dtype=bool,
                            chunks=(chunk_height, chunk_width))

    main_grp.create_dataset(labels_group, data=labs, dtype=np.uint32,
                            chunks=(chunk_height, chunk_width))

    main_grp.create_dataset(classes_group, data=np.array([[[label]]]),
                            dtype=np.uint32,
                            chunks=(1, 1, 1))

    return main_grp


def generate_zarr_group(dst_dir, image_specs):
    z_groups = generate_collection(**image_specs)

    return z_groups, None, None, None


def generate_zarr_array(dst_dir, image_specs):
    if dst_dir is None:
        z_groups = generate_collection(**image_specs)

        img_src = z_groups[image_specs["data_group"]]

        if "mask_group" in image_specs:
            mask_src = z_groups[image_specs["mask_group"]]
        else:
            mask_src = None

        if "labels_group" in image_specs:
            labels_src = z_groups[image_specs["labels_group"]]
        else:
            labels_src = None

        if "classes_group" in image_specs:
            classes_src = z_groups[image_specs["classes_group"]]
        else:
            classes_src = None

        return img_src, mask_src, labels_src, classes_src

    else:
        image_specs["data_group"] = "0"
        z_groups = generate_collection(**image_specs)
        image_specs["data_group"] = None

        img_idx = np.random.randint(10000)

        image_filename = f'{dst_dir}/zarr_array_{img_idx}.zarr'

        z_arr = zarr.open(image_filename, mode='w', shape=z_groups["0"].shape,
                          chunks=z_groups["0"].chunks)

        z_arr[:] = z_groups["0"][:]

        return image_filename, None, None, None


def generate_ndarray(dst_dir, image_specs):
    z_groups = generate_collection(**image_specs)

    img_src = z_groups[image_specs["data_group"]][:]

    if "mask_group" in image_specs:
        mask_src = z_groups[image_specs["mask_group"]][:]
    else:
        mask_src = None

    if "labels_group" in image_specs:
        labels_src = z_groups[image_specs["labels_group"]][:]
    else:
        labels_src = None

    if "classes_group" in image_specs:
        classes_src = z_groups[image_specs["classes_group"]][:]
    else:
        classes_src = None

    return img_src, mask_src, labels_src, classes_src


def generate_zarr_file(dst_dir, image_specs):
    z_groups = generate_collection(**image_specs)

    img_idx = np.random.randint(10000)

    image_filename = f'{dst_dir}/zarr_group_{img_idx}.zarr'
    z_grp_dst = zarr.open(image_filename, mode="w")
    zarr.copy_all(z_groups, z_grp_dst)

    return image_filename, None, None, None


def generate_tiffs(dst_dir, image_specs):
    src_data_group = image_specs["data_group"]

    if src_data_group is None:
        image_specs["data_group"] = "0"

    z_groups = generate_collection(**image_specs)

    img_idx = np.random.randint(10000)

    image_filename = f'{dst_dir}/img_{img_idx}.ome.tiff'
    with tifffile.TiffWriter(image_filename, bigtiff=True) as tif:
        tif.write(
            z_groups[image_specs["data_group"]][:],
            metadata={"axes": None},
            tile=(128, 128),
            photometric='rgb'
        )

        if "mask_group" in image_specs:
            tif.write(
                z_groups[image_specs["mask_group"]][:],
                metadata={"axes": "YX"},
                tile=(128, 128),
            )

        if "labels_group" in image_specs:
            tif.write(
                z_groups[image_specs["labels_group"]][:],
                metadata={"axes": "YX"},
                tile=(128, 128),
            )

        if "classes_group" in image_specs:
            tif.write(
                z_groups[image_specs["classes_group"]][:],
                metadata={"axes": "CYX"},
            )

    image_specs["data_group"] = src_data_group

    return image_filename, None, None, None


def generate_pngs(dst_dir, image_specs):
    z_groups = generate_collection(**image_specs)

    permute_order = zds.map_axes_order(source_axes=image_specs["source_axes"],
                                       target_axes="YXC")
    img_data = z_groups[image_specs["data_group"]][:]
    img_data = img_data.transpose(permute_order)
    img_data = img_data.reshape(-1, img_data.shape[-3], img_data.shape[-2],
                                img_data.shape[-1])
    img_data = img_data[0, ..., :3]

    image = Image.fromarray(img_data)

    img_idx = np.random.randint(10000)

    image_filename = f"{dst_dir}/img_{img_idx}.png"
    image.save(image_filename,
                quality_opts={'compress_level': 9, 'optimize': False})

    mask_data = z_groups[image_specs["mask_group"]][:]
    mask = Image.fromarray(mask_data)

    mask_filename = f"{dst_dir}/mask_{img_idx}.png"
    mask.save(mask_filename,
                quality_opts={'compress_level': 9, 'optimize': False})

    labels_data = z_groups[image_specs["labels_group"]][:]
    labels = Image.fromarray(labels_data)

    labels_filename = f"{dst_dir}/labels_{img_idx}.png"
    labels.save(labels_filename,
                quality_opts={'compress_level': 9, 'optimize': False})

    return image_filename, mask_filename, labels_filename, None


def generate_empty_file(dst_dir, image_specs):
    filename = Path(dst_dir) / image_specs["filename"]
    fp = open(filename, "w")
    fp.close()

    return filename, None, None, None


IMAGE_SPECS = [
    {
        "dst_dir": None,
        "source": "https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.1/6001237.zarr",
        "credit": ("© 2016-2023 University of Dundee & Open Microscopy "
                   "Environment. Creative Commons Attribution 4.0 "
                   "International License"),
        "specs": {
            "data_group": "0",
            "shape": [1, 4, 39, 1024, 1024],
            "chunks": [1, 1, 1, 1024, 1024],
            "source_axes": "TCZYX",
            "dtype": np.uint16
        }
    },
    {
        "dst_dir": None,
        "source": ("https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.4/idr0048A/"
                   "9846151.zarr/"),
        "credit": ("© 2016-2023 University of Dundee & Open Microscopy "
                   "Environment. Creative Commons Attribution 4.0 "
                   "International License"),
        "specs": {
            "data_group": "0/0",
            "shape": [1, 3, 1402, 5192, 2947],
            "chunks": [1, 1, 1, 1024, 1024],
            "source_axes": "TCZYX",
            "dtype": np.uint16
        }
    },
    {
        "dst_dir": "tests/test_zarrs",
        "source": generate_zarr_group,
        "credit": None,
        "specs": {
            "data_group": "0/0",
            "mask_group": "masks/0/0",
            "labels_group": "labels/0/0",
            "classes_group": "classes/0/0",
            "shape": [768, 512, 3],
            "chunks": [768, 512, 3],
            "source_axes": "YXC",
            "dtype": np.uint8
        }
    },
    {
        "dst_dir": "tests/test_zarrs",
        "source": generate_zarr_file,
        "credit": None,
        "specs": {
            "data_group": "0",
            "mask_group": "masks/0",
            "labels_group": "labels/0",
            "classes_group": "classes/0",
            "shape": [1, 4, 1125, 512],
            "chunks": [1, 4, 256, 128],
            "source_axes": "TCYX",
            "dtype": np.int32
        }
    },
    {
        "dst_dir": "tests/test_tiffs",
        "source": generate_tiffs,
        "credit": None,
        "specs": {
            "data_group": "0",
            "mask_group": "1",
            "labels_group": "2",
            "classes_group": "3",
            "shape": [1125, 512, 3],
            "chunks": [128, 128, 3],
            "source_axes": "YXC",
            "dtype": np.uint16
        }
    },
    {
        "dst_dir": "tests/test_tiffs",
        "source": generate_tiffs,
        "credit": None,
        "specs": {
            "data_group": "0/0",
            "mask_group": "0/1",
            "labels_group": "0/2",
            "classes_group": "0/3",
            "shape": [1125, 512, 3],
            "chunks": [128, 128, 3],
            "source_axes": "YXC",
            "dtype": np.uint16
        }
    },
    {
        "dst_dir": "tests/test_tiffs",
        "source": generate_tiffs,
        "credit": None,
        "specs": {
            "data_group": None,
            "shape": [1125, 512, 3],
            "chunks": [128, 128, 3],
            "source_axes": "YXC",
            "dtype": np.uint16
        }
    },
    {
        "dst_dir": "tests/test_tiffs",
        "source": generate_tiffs,
        "credit": None,
        "specs": {
            "data_group": 0,
            "shape": [1125, 512, 3],
            "chunks": [128, 128, 3],
            "source_axes": "YXC",
            "dtype": np.uint16
        }
    },
    {
        "dst_dir": "tests/test_pngs",
        "source": generate_pngs,
        "credit": None,
        "specs": {
            "data_group": "0",
            "mask_group": "masks/0",
            "labels_group": "labels/0",
            "classes_group": "classes/0",
            "shape": [1125, 512, 3],
            "chunks": [1125, 512, 3],
            "source_axes": "YXC",
            "dtype": np.uint8
        }
    },
    {
        "dst_dir": None,
        "source": generate_ndarray,
        "credit": None,
        "specs": {
            "data_group": "0",
            "mask_group": "masks/0",
            "labels_group": "labels/0",
            "classes_group": "classes/0",
            "shape": [1125, 3, 512],
            "chunks": [1125, 3, 512],
            "source_axes": "YCX",
            "dtype": np.uint8
        }
    },
    {
        "dst_dir": None,
        "source": generate_zarr_array,
        "credit": None,
        "specs": {
            "data_group": "0",
            "mask_group": "masks/0",
            "labels_group": "labels/0",
            "classes_group": "classes/0",
            "shape": [1, 3, 1024, 512, 1],
            "chunks": [1, 3, 256, 128, 1],
            "source_axes": "ZCXYT",
            "dtype": np.float32
        }
    },
    {
        "dst_dir": "tests/test_zarrs",
        "source": generate_zarr_array,
        "credit": None,
        "specs": {
            "data_group": None,
            "shape": [1, 3, 1024, 1024, 1],
            "chunks": [1, 3, 256, 256, 1],
            "source_axes": "ZCXYT",
            "dtype": np.float32
        }
    },
    {
        "dst_dir": None,
        "source": ("https://cellpainting-gallery.s3.amazonaws.com/"
                   "cpg0000-jump-pilot/source_4/images/2020_11_04_CPJUMP1/"
                   "images/BR00116991__2020-11-05T19_51_35-Measurement1/"
                   "Images/r01c01f01p01-ch1sk1fk1fl1.tiff"),
        "credit": ("From the dataset cpg0000 (Chandrasekaran et al., 2022), "
                   "available from the Cell Painting Gallery (Weisbart et al.,"
                   " 2024) on the Registry of Open Data on AWS "
                   "(https://registry.opendata.aws/cellpainting-gallery/)"),
        "specs": {
            "data_group": None,
            "shape": [1080, 1080],
            "chunks": [1080, 1080],
            "source_axes": "YX",
            "dtype": np.uint8
        }
    },
]

MASKABLE_IMAGE_SPECS = [
    {
        "dst_dir": None,
        "source": ("https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.4/idr0073A/"
                   "9798462.zarr"),
        "credit": ("© 2016-2023 University of Dundee & Open Microscopy "
                   "Environment. Creative Commons Attribution 4.0 "
                   "International License"),
        "specs": {
            "data_group": "0",
            "shape": [1, 3, 1, 16433, 21115],
            "chunks": [1, 1, 1, 1024, 1024],
            "source_axes": "TCZYX",
            "axes": "YXC",
            "roi": "(0,0,0,0,0):(1,-1,1,-1,-1)",
            "dtype": np.uint16
        }
    },
]


UNSUPPORTED_IMAGE_SPECS = [
    {
        "dst_dir": "tests/test_unsupported",
        "source": generate_empty_file,
        "credit": None,
        "specs": {
            "filename": "image.unknown",
            "wrong_data_group": None,
        }
    },
    {
        "dst_dir": "tests/test_unsupported",
        "source": generate_empty_file,
        "credit": None,
        "specs": {
            "filename": "image.txt",
            "wrong_data_group": None,
        }
    },
    {
        "dst_dir": "tests/test_unsupported",
        "source": generate_empty_file,
        "credit": None,
        "specs": {
            "filename": "image.zip",
            "wrong_data_group": None,
        }
    },
    {
        "dst_dir": "tests/test_zarrs",
        "source": generate_zarr_group,
        "credit": None,
        "specs": {
            "data_group": "0",
            "shape": [1, 2, 2],
            "chunks": [1, 2, 2],
            "source_axes": "CYX",
            "dtype": np.uint8,
            "wrong_data_group": None,
        }
    },
    {
        "dst_dir": "tests/test_zarrs",
        "source": generate_zarr_file,
        "credit": None,
        "specs": {
            "data_group": "0",
            "shape": [1, 2, 2],
            "chunks": [1, 2, 2],
            "source_axes": "CYX",
            "dtype": np.uint8,
            "wrong_data_group": None,
        }
    },
    {
        "dst_dir": "tests/test_tiffs",
        "source": generate_tiffs,
        "credit": None,
        "specs": {
            "data_group": "0",
            "shape": [2, 2, 3],
            "chunks": [2, 2, 3],
            "source_axes": "YXC",
            "dtype": np.uint8,
            "wrong_data_group": dict(data_group="0"),
        }
    },
]



import pytest

import zarrdataset as zds

import numpy as np


@pytest.fixture(scope="function")
def input_image():
    np.random.seed(75146)
    shape = [np.random.randint(1, 256) if ax != "C" else 3
             for ax in "YXC"]
    img = np.random.randint(low=0, high=255, size=shape, dtype=np.uint8)
    axes = "YXC"
    shape = dict(
        (ax, s)
        for ax, s in zip(axes, shape)
    )
    return img, shape, axes


def test_MaskGenerator(input_image):
    mask_generator = zds.MaskGenerator(axes=None)

    img, _, _ = input_image

    with pytest.raises(NotImplementedError):
        _ = mask_generator(img)


@pytest.mark.parametrize("mask_scale, min_size, area_threshold, thresh, axes",
[
    (0.5, 5, 10, None, "YX"),
    (2, 5, 10, 0.5, "ZYX"),
])
def test_WSITissueMaskGenerator(input_image, mask_scale, min_size,
                                area_threshold,
                                thresh,
                                axes):
    img, img_shape, img_axes = input_image

    tissue_mask_generator = zds.WSITissueMaskGenerator(
        mask_scale=mask_scale,
        min_size=min_size,
        area_threshold=area_threshold,
        thresh=thresh,
        axes=axes
    )

    mask = tissue_mask_generator(img)

    assert isinstance(mask, np.ndarray), \
        (f"Expected mask to be a Numpy NDArray, got {type(mask)} instead.")

    expected_shape = tuple(
        round(img_shape[ax] * mask_scale) if ax in img_axes else 1
        for ax in tissue_mask_generator.axes
    )

    assert mask.shape == expected_shape, \
        (f"Expected mask has shape {expected_shape}, got {mask.shape} "
         f"instead.")



import zarrdataset as zds
from unittest import mock
import pytest
import importlib
import os
import shutil
import tifffile
import random

from pathlib import Path
import zarr
import numpy as np
from tests.utils import IMAGE_SPECS, UNSUPPORTED_IMAGE_SPECS


@pytest.fixture
def dummy_array():
    source_data = np.ones(shape=(16, 16, 3), dtype=np.uint8)
    source_axes = "YXC"
    return source_data, source_axes


@pytest.fixture(scope="function")
def input_image(request):
    if isinstance(request.param["source"], str):
        yield request.param["source"], request.param["specs"]

    else:
        dst_dir = request.param["dst_dir"]

        if dst_dir is not None:
            dst_dir = Path(request.param["dst_dir"])
            dst_dir.mkdir(parents=True, exist_ok=True)

        yield (request.param["source"](request.param["dst_dir"],
                                       request.param["specs"])[0],
               request.param["specs"])

        if dst_dir is not None and os.path.isdir(dst_dir):
            shutil.rmtree(dst_dir)


@pytest.mark.parametrize("input_image", IMAGE_SPECS, indirect=["input_image"])
def test_image_formats(input_image):
    image_src, image_specs = input_image

    expected_shape = tuple(image_specs["shape"])
    expected_chunks = tuple(image_specs["chunks"])

    store = None

    try:
        arr, store = zds.image2array(
            image_src,
            data_group=image_specs["data_group"]
        )

        assert arr.shape == expected_shape, \
            (f"Expected image of shape {expected_shape}, got {arr.shape}")

        assert arr.chunks == expected_chunks, \
            (f"Expected chunks {expected_chunks}, got {arr.chunks}")

    finally:
        if store is not None:
            store.close()


@pytest.mark.parametrize("input_image", UNSUPPORTED_IMAGE_SPECS,
                         indirect=["input_image"])
def test_unsupported_image_formats(input_image):
    image_src, image_specs = input_image

    with pytest.raises(ValueError):
        _ = zds.image2array(image_src,
                            data_group=image_specs["wrong_data_group"])


@pytest.mark.parametrize("shape, chunk_size, source_axes, expected_chunk_size",
    [
        ([16, 16, 3], [8, 8, 2], "YXC", [8, 8, 2]),
        ([16, 16, 3], None, "YXC", [16, 16, 3]),
        ([16, 16, 3], None, "YXC", [16, 16, 3]),
    ]
)
def test_ImageBase(shape, chunk_size, source_axes, expected_chunk_size):
    img = zds.ImageBase(shape, chunk_size=chunk_size, source_axes=source_axes,
                        mode="images")
    assert img.shape == shape, \
        (f"Expected shape is {shape}, got {img.shape} instead")

    assert img.chunk_size == expected_chunk_size, \
        (f"Expected chunk size is {expected_chunk_size}, got {img.chunk_size}"
         f" instead")

    expected_scale = dict((ax, 1) for ax in source_axes)
    assert img.scale == expected_scale, \
        (f"Expected scale before rescaling to be {expected_scale}, "
         f"got {img.scale} instead")


def test_ImageBase_slicing():
    shape = (16, 16, 3)
    axes = "YXC"
    ref_shape = (8, 64, 90)
    ref_axes = "ZXY"
    img = zds.ImageBase(shape, chunk_size=None, source_axes=axes, mode="image")

    img.rescale(spatial_reference_shape=ref_shape,
                spatial_reference_axes=ref_axes)
    
    ref_shape_dict = dict(
        (ax, s)
        for ax, s in zip(ref_axes, ref_shape)
    )

    expected_scales = dict(
        (ax, (s / ref_shape_dict[ax]) if ax in ref_axes else 1.0)
        for ax, s in zip(axes, shape)
    )

    assert img.scale == expected_scales, \
        (f"Expected scale after rescaling the image to be {expected_scales}, "
         f"got {img.scale} instead.")

    rescaled_img_shape = [
        (ref_shape_dict[ax] * img.scale[ax]) if ax in ref_axes else s
        for ax, s in zip(img.axes, img.shape)
    ]

    assert all([int(s) == se for s, se in zip(rescaled_img_shape, shape)]), \
        (f"Expected selection slice(None), None, :, to be of shape {shape}, "
         f"got {rescaled_img_shape} instead.")

    assert img[:].shape == shape, \
        (f"Expected selection slice(None), None, :, to be of shape {shape}, "
         f"got {img[:].shape} instead.")

    random.seed(44512)
    selection_1 = dict(
        (ax, slice(0, random.randint(1, r_s)))
        for ax, r_s in zip(ref_axes, ref_shape)
    )

    expected_selection_shape = tuple(
        round(expected_scales[ax] * selection_1[ax].stop)
        if ax in ref_axes else s
        for ax, s in zip(axes, shape)
    )

    img_sel_1 = img[selection_1]

    assert img_sel_1.shape == expected_selection_shape, \
        (f"Expected selection {selection_1} to have shape "
         f"{expected_selection_shape}, got {img_sel_1.shape} instead")

    selection_2 = tuple(selection_1[ax] for ax in ref_axes)

    img_sel_2 = img[selection_2]

    assert img_sel_2.shape == expected_selection_shape, \
        (f"Expected selection {selection_2} to have shape "
         f"{expected_selection_shape}, got {img_sel_2.shape} instead")


def test_ImageBase_padding():
    shape = (16, 16, 3)
    axes = "YXC"
    img = zds.ImageBase(shape, chunk_size=None, source_axes=axes, mode="image")

    random.seed(44512)
    selection_1 = dict(
        (ax, slice(random.randint(-10, 0),
                   random.randint(1, r_s + 10)))
        for ax, r_s in zip(axes, shape)
    )

    expected_selection_shape = tuple(
        selection_1[ax].stop - selection_1[ax].start for ax in axes
    )

    img_sel_1 = img[selection_1]

    assert img_sel_1.shape == expected_selection_shape, \
        (f"Expected selection {selection_1} to have shape "
         f"{expected_selection_shape}, got {img_sel_1.shape} instead")


@pytest.mark.parametrize("axes, roi, expected_size", [
    (None, None, (16, 16, 3)),
    (None, slice(None), (16, 16, 3)),
    (None, (slice(2, None, None), slice(7, None, None), slice(1, 2, None)),
     (14, 9, 1)),
    (None, [slice(2, None, None), slice(7, None, None), slice(1, 2, None)],
     (14, 9, 1)),
    (None, "(10,2,0):(5,10,-1)", (5, 10, 3)),
    ("WYC", "(7,8,0):(5,1,-1)", (1, 5, 3)),
])
def test_ImageLoader_supported_rois(dummy_array, axes, roi, expected_size):
    source_data, source_axes = dummy_array

    img = zds.ImageLoader(source_data, source_axes, axes=axes, roi=roi)

    assert img[:].shape == expected_size, \
        (f"Expected size of image ROI is {expected_size}, got {img[:].shape} "
         f"instead.")


@pytest.mark.parametrize("axes, roi", [
    (None, (slice(None), 10, slice(1, 2, None))),
    ("YX", slice(0, 10)),
    ("YX", (slice(1, 2, None), )),
    (None, {"Y": slice(None), "X": slice(None)}),
    ("YC", None),
])
def test_ImageLoader_unsupported_rois(dummy_array, axes, roi):
    source_data, source_axes = dummy_array

    with pytest.raises(ValueError):
        img = zds.ImageLoader(source_data, source_axes, axes=axes, roi=roi)


@pytest.mark.parametrize("axes, permute_order", [
    ("YXC", [0, 1, 2]),
    ("CXY", [2, 1, 0]),
    ("WZCYX", [2, 0, 1])
])
def test_ImageLoader_axes(dummy_array, axes, permute_order):
    source_data, source_axes = dummy_array

    img = zds.ImageLoader(source_data, source_axes, axes=axes)

    assert img.permute_order == permute_order, \
        (f"Incorrect permute ordering, expected {permute_order}, got "
         f"{img.permute_order} instead.")


def test_ImageLoader_function(dummy_array):
    class TestFunc(zds.MaskGenerator):
        def __init__(self):
            super(TestFunc, self).__init__(axes="YX")

        def _compute_transform(self, image):
            return np.mean(image, axis=-1)

    source_data, source_axes = dummy_array

    img = zds.ImageLoader(source_data, source_axes, axes="YXC",
                          image_func=TestFunc())

    res = img[:]

    assert res.shape == (16, 16), \
        (f"Expected shape of transformed image to be (16, 16), got {res.shape}"
         f" instead.")


@pytest.mark.parametrize("input_image", IMAGE_SPECS, indirect=["input_image"])
def test_ImageLoader_formats(input_image):
    image_src, image_specs = input_image

    expected_shape = image_specs["shape"]

    img = zds.ImageLoader(image_src,
                          source_axes=image_specs["source_axes"],
                          data_group=image_specs["data_group"])

    assert img.shape == expected_shape, \
        (f"Expected image of shape {expected_shape}, got {img.shape}")


def test_ImageCollection():
    img = zarr.array(data=np.random.randn(15, 15), shape=(15, 15),
                     chunks=(5, 5))
    mask = zarr.array(data=np.random.randn(10, 10), shape=(10, 10),
                      chunks=(5, 5))

    image_collection = zds.ImageCollection(collection_args=dict(
            images=dict(
                filename=img,
                data_group=None,
                source_axes="YX",
                axes=None,
                roi=None
            ),
            masks=dict(
                filename=mask,
                data_group=None,
                source_axes="YX",
                axes=None,
                roi=None
            )
        )
    )

    assert np.array_equal(image_collection.collection["masks"][:], mask[:]),\
        (f"Original mask and mask in collection are not the same. Expected "
         f"{mask[0, 0]}, got {image_collection.collection['masks'][0, 0]}")

    image_collection = zds.ImageCollection(collection_args=dict(
            images=dict(
                filename=img,
                data_group=None,
                source_axes="YX",
                axes=None,
                roi=None
            )
        )
    )

    expected_mask_size = [s//c for s, c in zip(img.shape, img.chunks)]

    assert image_collection.collection["masks"].shape == expected_mask_size,\
        (f"Auto-generated mask is not the same size as of the image chunks")
    
    patch = image_collection[:]

    assert isinstance(patch, dict), \
        (f"Expected patch extracted from image collection to be a tuple, got "
         f"{type(patch)} instead.")

    assert patch["images"].shape == img.shape, \
        (f"Expected patch shape to be {img.shape}, got {patch['images'].shape}"
         f" instead.")


def test_compatibility_no_tifffile():
    with mock.patch.dict('sys.modules', {'tifffile': None}):
        importlib.reload(zds._imageloaders)

        assert not zds._imageloaders.TIFFFILE_SUPPORT,\
            (f"If TiffFile is not installed, image loading functionalities "
             f"depending on it should be disabled, but TIFFFILE_SUPPORT is "
             f"{zds._imageloaders.TIFFFILE_SUPPORT}")

    with mock.patch.dict('sys.modules', {'tifffile': tifffile}):
        importlib.reload(zds._imageloaders)
        importlib.reload(zds)



import pytest
import operator

import zarrdataset as zds

import shutil
from pathlib import Path
import zarr
import boto3


@pytest.mark.parametrize("rois_str, expected_rois", [
    ("", [slice(None)]),
    ([""], [slice(None)]),
    ("(0):(10)", [(slice(0, 10), )]),
    ("(2, 5):(-1, 6)", [(slice(2, None), slice(5, 11))]),
])
def test_parse_rois(rois_str, expected_rois):
    parsed_rois = zds.parse_rois(rois_str)

    assert all(map(operator.eq, parsed_rois, expected_rois)), \
        (f"Expected parsed ROIs to be {expected_rois}, got {parsed_rois} "
         f"instead")


@pytest.mark.parametrize("filename, default_source_axes, default_data_group,"
                         "default_axes, default_rois, override_meta,"
                         "expected_metadata", [
    ("test.png", "YX", "0/0", None, None, True, [
        dict(
            filename="test.png",
            source_axes="YX",
            axes="YX",
            data_group="0/0",
            roi=None)
    ]),
    ("file; test.png", "CYX", "", "WCYX", None, True, [
        dict(
            filename="file; test.png",
            source_axes="CYX",
            axes="WCYX",
            data_group="",
            roi=None)
    ]),
    ("test.png;0/0;CYX:YXC;(0,0,1):(-1,-1,3)", "YCX", None, None,
     [(slice(0, 2), slice(0, 2), slice(None)),
      (slice(None), slice(None), slice(None))], False, [
        dict(
            filename="test.png",
            source_axes="CYX",
            axes="YXC",
            data_group="0/0",
            roi=(slice(None), slice(None), slice(1, 4))
        ),
    ]),
    ("test.png;0/0;CYX:YXC;(0,0,1):(-1,-1,3)", "YCX", None, None,
     [(slice(0, 2), slice(0, 2), slice(None)),
      (slice(None), slice(None), slice(None))], True, [
        dict(
            filename="test.png",
            source_axes="YCX",
            axes="YCX",
            data_group="0/0",
            roi=(slice(0, 2), slice(0, 2), slice(None))
        ),
        dict(
            filename="test.png",
            source_axes="YCX",
            axes="YCX",
            data_group="0/0",
            roi=(slice(None), slice(None), slice(None))
        ),
    ]),
    ("test.png;0/0;CYX:YXC;(0,0,1):(-1,-1,3);(2,3,1):(2,2,-1)", None, None,
     None, None, False, [
        dict(
            filename="test.png",
            source_axes="CYX",
            axes="YXC",
            data_group="0/0",
            roi=(slice(None), slice(None), slice(1, 4))
        ),
        dict(
            filename="test.png",
            source_axes="CYX",
            axes="YXC",
            data_group="0/0",
            roi=(slice(2, 4), slice(3, 5), slice(1, None))
        )
    ]),
    ([0, 1, 2], "C", "", None, None, True, [
        dict(
            filename=[0, 1, 2],
            source_axes="C",
            axes="C",
            data_group="",
            roi=None)
        ]
    ),
])
def test_parse_metadata(filename, default_source_axes, default_data_group,
                        default_axes,
                        default_rois,
                        override_meta,
                        expected_metadata):
    parsed_metadata = zds.parse_metadata(filename, default_source_axes,
                                         default_data_group,
                                         default_axes,
                                         default_rois,
                                         override_meta)
    
    assert isinstance(parsed_metadata, list), \
        (f"Expected parsed metadata to be a list of dictionaries, got "
         f"{type(parsed_metadata)}")

    assert all(all(par_meta[ax] == exp_val for ax, exp_val in exp_meta.items())
               for par_meta, exp_meta in zip(parsed_metadata,
                                             expected_metadata)), \
        (f"Expected parsed metadata to be {expected_metadata}, got "
         f"{parsed_metadata} instead.")


@pytest.mark.parametrize("source_axes, target_axes, expected_order", [
    ("YX", "XY", [1, 0]),
])
def test_map_axes_order(source_axes, target_axes, expected_order):
    transpose_order = zds.map_axes_order(source_axes, target_axes)

    assert transpose_order == expected_order, \
        (f"Expected order to be {expected_order}, got {transpose_order} "
         f"instead")


@pytest.mark.parametrize("source_axes, axes_selection, expected_sel_slices, "
                         "expected_unfixed_axes", [
    ("YX", dict(Y=slice(None), X=slice(None)), (slice(None), slice(None)),
     "YX"),
    ("Y", dict(Y=None), (slice(None), ), "Y"),
])
def test_select_axes(source_axes, axes_selection, expected_sel_slices,
                     expected_unfixed_axes):
    sel_slices, unfixed_axes = zds.select_axes(source_axes, axes_selection)

    assert sel_slices == expected_sel_slices, \
        (f"Expected selection to be {expected_sel_slices}, got {sel_slices} "
         f"instead.")

    assert unfixed_axes == expected_unfixed_axes, \
        (f"Expected unfixed axes to be {expected_unfixed_axes}, got "
         f"{unfixed_axes} instead.")


@pytest.mark.parametrize("filename_sample, expected_s3_obj", [
    ("test.png", None),
    ("https://live.staticflickr.com/4908/31072787307_59f7943caa_o.jpg", dict(
        bucket_name="4908",
        endpoint_url="https://live.staticflickr.com",
        s3=boto3.client("s3", aws_access_key_id="", aws_secret_access_key="",
                        region_name="us-east-2",
                        endpoint_url="https://live.staticflickr.com"))),
])
def test_connect_s3(filename_sample, expected_s3_obj):
    s3_obj = zds.connect_s3(filename_sample)

    if expected_s3_obj is None:
        assert s3_obj is None, \
            (f"Expected connection to be None when passing a local filename.")
    else:
        assert s3_obj["bucket_name"] == expected_s3_obj["bucket_name"], \
            (f"Expected bucket name {expected_s3_obj['bucket_name']}, got "
             f"{s3_obj['bucket_name']} instead.")
        assert s3_obj["endpoint_url"] == expected_s3_obj["endpoint_url"], \
            (f"Expected endpoiint url {expected_s3_obj['endpoint_url']}, got "
             f"{s3_obj['endpoint_url']} instead.")
        assert (s3_obj["s3"]._endpoint.host
                == expected_s3_obj["s3"]._endpoint.host), \
            (f"Expected client endpoint {expected_s3_obj['s3']._endpoint.host}"
             f", got {s3_obj['s3']._endpoint.host} instead.")


def test_consolidated():
    root_path = Path("tests/test_zarrs")
    root_path.mkdir(parents=True, exist_ok=True)

    try:
        root = zarr.open(root_path, mode="w")
        root.create_dataset("test/test_arr", shape=10, dtype="i8")

        zarr.consolidate_metadata(root_path)

        isconsolidated = zds.isconsolidated(str(root_path))

        assert isconsolidated, \
            (f"Expected {root_path} be consolidated, got {isconsolidated} "
             f"instead.")

    finally:
        shutil.rmtree(root_path)


@pytest.mark.parametrize("arr_src, expected_consolidated", [
    ("test.zarr", False),
    ("https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.1/6001237.zarr", False),
])
def test_no_consolidated(arr_src, expected_consolidated):
    isconsolidated = zds.isconsolidated(arr_src)

    assert isconsolidated == expected_consolidated, \
        (f"Expected {arr_src} be consolidated={expected_consolidated}, got"
         f"{isconsolidated} instead.")


@pytest.mark.parametrize("selection_range, scale, expected_coords", [
    ([slice(None), slice(None)], [1, 1], (slice(None), slice(None))),
    ([slice(None), slice(None)], 1, (slice(None), slice(None))),
    ([None, slice(None)], 1, (slice(None), slice(None))),
    ([slice(2, 4)], 0.5, (slice(1, 2), )),
    ([(1, 5)], 2.01, (slice(2, 11), )),
    ([range(4)], 2, (slice(0, 8, 1), )),
    ((8, ), 0.25, (slice(2, 3), )),
])
def test_scale_coords(selection_range, scale, expected_coords):
    scaled_coords = zds.scale_coords(selection_range, scale)
    assert scaled_coords == expected_coords, \
        (f"Expected scaled coords {expected_coords}, got {scaled_coords} "
         f"instead.")


@pytest.mark.parametrize("selection_range, scale", [
    ([dict(Y=slice(None))], [1]),
])
def test_unsupported_scale_coords(selection_range, scale):
    with pytest.raises(ValueError):
        scaled_coords = zds.scale_coords(selection_range, scale)


@pytest.mark.parametrize("index, roi, source_axes, axes, expected_roi", [
    (dict(Y=slice(None), X=slice(None)), (slice(None), slice(None)), "YX",
     "YX", (slice(None), slice(None))),
    (dict(Y=slice(0, 5)), (slice(None), ), "Y", "Y", (slice(0, 5), )),
    (dict(Y=slice(0, 5)), (slice(5, 10), ), "Y", "Y", (slice(5, 10), )),
    (dict(Y=slice(0, 5)), (slice(5, 7), ), "Y", "Y", (slice(5, 7), )),
    (dict(Y=slice(0, 5)), (slice(None), slice(None)), "YZ", "Y",
     (slice(0, 5), slice(None))),
    (dict(W=slice(0, 5)), (slice(None), slice(None)), "YZ", "Z",
     (slice(None), slice(None))),
])
def test_translate2roi(index, roi, source_axes, axes, expected_roi):
    translated_roi = zds.translate2roi(index, roi, source_axes, axes)

    assert translated_roi == expected_roi, \
        (f"Expected translated ROI to be {expected_roi}, got {translated_roi} "
         f"instead.")



import pytest
import os
import shutil
from pathlib import Path
from unittest import mock
from collections import OrderedDict
from _collections_abc import dict_items

import operator

import zarrdataset as zds
import importlib
import tqdm
import numpy as np

from tests.utils import IMAGE_SPECS

import torch
from torch.utils.data import DataLoader, ChainDataset


def input_target_dtype_fn(input, target):
    return input.astype(np.float64), target.astype(np.float64)


@pytest.fixture(scope="function")
def image_dataset_specs(request):
    if not isinstance(request.param, list):
        params = [request.param]
    else:
        params = request.param

    dst_dirs = []

    filenames = []
    labels_filenames = []
    mask_filenames = []
    specs = []

    data_group = None
    source_axes = None
    mask_group = None
    labels_group = None

    for par in params:
        if isinstance(par["source"], str):
            dst_dirs.append(None)

        else:
            dst_dirs.append(par["dst_dir"])

        if dst_dirs[-1] is not None:
            dst_dir = Path(par["dst_dir"])
            dst_dir.mkdir(parents=True, exist_ok=True)

        (image_filename,
         mask_filename,
         labels_filename,
         _) = par["source"](par["dst_dir"], par["specs"])

        data_group = par["specs"]["data_group"]
        source_axes = par["specs"]["source_axes"]
        filenames.append(image_filename)

        if mask_filename is not None:
            mask_filenames.append(mask_filename)
            mask_group = par["specs"]["mask_group"]

        if labels_filename is not None:
            input_label_transform = input_target_dtype_fn
            target_transform = zds.ToDtype(np.int64)
            labels_filenames.append(labels_filename)
            labels_group = par["specs"]["labels_group"]

        specs.append(par["specs"])

    dataset_specs = [
        zds.ImagesDatasetSpecs(
            filenames=filenames,
            data_group=data_group,
            source_axes=source_axes,
        )
    ]

    if mask_filenames:
        dataset_specs.append(
            zds.MasksDatasetSpecs(
                filenames=mask_filenames,
                data_group=mask_group,
                source_axes="YX",
            )
        )

    if labels_filenames:
        dataset_specs[0]["transforms"].append(
            (("images", ), zds.ToDtype(np.float32))
        )

        dataset_specs.append(
            zds.LabelsDatasetSpecs(
                filenames=labels_filenames,
                data_group=labels_group,
                source_axes="YX",
                transform=target_transform,
                input_label_transform=input_label_transform,
            )
        )

    if len(dataset_specs) == 1:
        dataset_specs = dataset_specs[0]

    yield dataset_specs, specs[0]

    for dst_dir in dst_dirs:
        if dst_dir is not None and os.path.isdir(dst_dir):
            shutil.rmtree(dst_dir)


@pytest.fixture(scope="function")
def patch_sampler_specs(request):
    patch_size, allow_incomplete_patches = request.param
    patch_sampler = zds.PatchSampler(
        patch_size=patch_size,
        allow_incomplete_patches=allow_incomplete_patches
    )
    return patch_sampler, patch_size, allow_incomplete_patches



@pytest.mark.parametrize("image_dataset_specs", [
    IMAGE_SPECS[10],
], indirect=["image_dataset_specs"])
def test_compatibility_no_tqdm(image_dataset_specs):
    with mock.patch.dict('sys.modules', {'tqdm': None}):
        importlib.reload(zds._zarrdataset)

        assert isinstance(object, type(zds._zarrdataset.tqdm).__bases__), \
            (f"When `tqdm` is not installed, progress bars should be disabled,"
             f" and a dummy interface derived from object should be used by "
             f"ZarrDataset. Got an interface derived from "
             f"{type(zds._zarrdataset.tqdm).__bases__} instead")

        dataset_specs, _ = image_dataset_specs

        dataset = zds._zarrdataset.ZarrDataset(
            dataset_specs=dataset_specs,
            progress_bar=True
        )

        try:
            next(iter(dataset))

        except Exception as e:
            raise AssertionError(f"No exceptions where expected, got {e} "
                                 f"instead.")

    with mock.patch.dict('sys.modules', {'tqdm': tqdm}):
        importlib.reload(zds._zarrdataset)
        importlib.reload(zds)


@pytest.mark.parametrize("dataset_spec_class", [
    zds.DatasetSpecs,
    zds.ImagesDatasetSpecs,
    zds.MasksDatasetSpecs,
    zds.LabelsDatasetSpecs
])
def test_DataSpecs(dataset_spec_class):
    generic_specs = dataset_spec_class(
        modality="generic",
        filenames="",
        source_axes=""
    )

    assert generic_specs["modality"] == "generic", \
        (f"Expected mode to be `generic`, as assigned during initialization of"
         f" the dataset specification, got {generic_specs['modality']} "
         f"instead.")

    assert generic_specs["filenames"] == "", \
        (f"Expected filenames to be ````, as assigned during initialization of"
         f" the dataset specification, got {generic_specs['filenames']} "
         f"instead.")

    assert isinstance(generic_specs["transforms"], list), \
        (f"Expected transforms to be an ordered dicionary, got "
         f"{type(generic_specs['transforms'])} instead.")

    assert isinstance(generic_specs.items(), dict_items), \
        (f"Expected DatasetSpecs to return `dict_items` when called `.items()`"
         f", got {type(generic_specs.items())} instead.")

    ds = zds.ZarrDataset()
    try:
        ds.add_modality(**generic_specs)

    except Exception as e:
        raise AssertionError(f"No exceptions where expected, got {e} "
                             f"instead.")


def test_no_images_ZarrDataset():
    ds = zds.ZarrDataset()

    with pytest.raises(ValueError):
        ds_iter = iter(ds)
        sample = next(ds_iter)


@pytest.mark.parametrize("image_dataset_specs", [
    IMAGE_SPECS[10],
], indirect=["image_dataset_specs"])
def test_string_ZarrDataset(image_dataset_specs):
    dataset_specs, specs = image_dataset_specs

    patch_sampler = zds.PatchSampler(0)
    ds = zds.ZarrDataset(dataset_specs,
                         patch_sampler=patch_sampler)

    ds_str_repr = str(ds)
    assert (ds_str_repr.startswith("ZarrDataset")
            and all([mode['modality'] in ds_str_repr
                     for mode in dataset_specs])
            and f"Using {dataset_specs[0]['modality']}" in ds_str_repr
            and f"{patch_sampler}" in ds_str_repr), \
        (f"Spected string representation start with `ZarrDataset`, list all "
         f"modalities, and state what modality is being used as reference. "
         f"Got `{ds_str_repr}` instead.")


@pytest.mark.parametrize(
    "image_dataset_specs, transforms", [
        (IMAGE_SPECS[6], (("images", ), zds.ToDtype(np.float64))),
        (IMAGE_SPECS[6], [(("images", ), zds.ToDtype(np.float32)),
                          (("images", ), zds.ToDtype(np.float64))]),
        (IMAGE_SPECS[6], [(("images", ), [zds.ToDtype(np.float32),
                                          zds.ToDtype(np.float64)])]),
    ],
    indirect=["image_dataset_specs"]
)
def test_ZarrDataset_transforms(image_dataset_specs, transforms):
    dataset_specs, _ = image_dataset_specs

    ds = zds.ZarrDataset(dataset_specs=dataset_specs)

    if not isinstance(dataset_specs, list):
        dataset_specs = [dataset_specs]

    if not isinstance(transforms, list):
        transforms = [transforms]

    expected_transforms_per_mode = {}
    for mode, mode_transform in transforms:
        expected_transforms_per_mode[mode] =\
            expected_transforms_per_mode.get(mode, 0) + 1

        ds.add_transform(mode, mode_transform)

    transforms_per_mode = {}
    for mode, mode_transform in ds._transforms:
        transforms_per_mode[mode] = transforms_per_mode.get(mode, 0) + 1

    assert all(map(lambda mode_n_transforms:
                   transforms_per_mode.get(mode_n_transforms[0], 0)
                   == mode_n_transforms[1],
                   expected_transforms_per_mode.items())), \
           (f"Expected number of transforms to be "
            f"{expected_transforms_per_mode}, got "
            f"{transforms_per_mode} instead.")

    for sample in ds:
        if any(["labels" in mode["modality"] for mode in dataset_specs]):
            sample_array = sample[0]

        else:
            sample_array = sample

        assert sample_array.dtype == np.float64, \
            (f"Sample data type should be numpy.float64, got "
             f"{sample_array.dtype} instead.")


@pytest.mark.parametrize(
    "image_dataset_specs, shuffle, return_positions, return_worker_id", [
        (IMAGE_SPECS[10], False, False, False),
        (IMAGE_SPECS[10], True, True, True),
        (IMAGE_SPECS[11], False, False, False),
        (IMAGE_SPECS[11], False, False, False),
    ],
    indirect=["image_dataset_specs"]
)
def test_ZarrDataset(image_dataset_specs, shuffle, return_positions,
                     return_worker_id):
    dataset_specs, specs = image_dataset_specs

    ds = zds.ZarrDataset(
        dataset_specs=dataset_specs,
        return_positions=return_positions,
        return_worker_id=return_worker_id,
        shuffle=shuffle,
    )

    if not isinstance(dataset_specs, list):
        dataset_specs = [dataset_specs]

    ds.add_transform(dataset_specs[0]["modality"], zds.ToDtype(np.float64))

    array_idx = 0
    label_idx = 1

    if return_positions:
        array_idx += 1
        label_idx += 1

    if return_worker_id:
        array_idx += 1
        label_idx += 1

    n_samples = 0

    for sample in ds:
        n_samples += 1
        if (any(["labels" in mode["modality"] for mode in dataset_specs])
           or return_positions
           or return_worker_id):
            assert isinstance(sample, tuple), \
                (f"When `return_positions`, `return_worker_id` or a labels "
                 f"dataset specification is passed to ZarrDataset, retrieved "
                 f"samples should be a tuple, got {type(sample)} instead.")

            sample_array = sample[array_idx]

        else:
            sample_array = sample

        assert isinstance(sample_array, np.ndarray), \
            (f"Sample should be a Numpy NDArray, got {type(sample_array)}"
             f" instead.")

        assert sample_array.dtype == np.float64, \
            (f"Sample data type should be numpy.float64, got "
             f"{sample_array.dtype} instead.")

        assert tuple(sample_array.shape) == tuple(specs["shape"]), \
            (f"Sample expected to have shape {tuple(specs['shape'])}, got "
             f"{sample_array.shape} ({sample_array.dtype}) instead")

        if "labels" in dataset_specs:
            expected_labels_shape = tuple(
                specs["shape"][specs["source_axes"].index(ax)]
                for ax in "YX"
            )

            labels_array = sample[label_idx]

            assert tuple(labels_array.shape) == expected_labels_shape, \
                (f"Labels expected to have shape {expected_labels_shape}, got "
                f"{labels_array.shape} ({labels_array.dtype}) instead")

    assert n_samples > 0, ("Expected more than zero samples extracted from "
                           "this experiment.")


@pytest.mark.parametrize(
    "image_dataset_specs, patch_sampler_specs, shuffle, draw_same_chunk", [
        (IMAGE_SPECS[10], (32, False), True, False),
        (IMAGE_SPECS[10], (32, False), True, True),
        (IMAGE_SPECS[10], (32, False), False, True),
    ],
    indirect=["image_dataset_specs", "patch_sampler_specs"]
)
def test_patched_ZarrDataset(image_dataset_specs, patch_sampler_specs,
                             shuffle,
                             draw_same_chunk):
    dataset_specs, specs = image_dataset_specs
    patch_sampler, patch_size, allow_incomplete_patches = patch_sampler_specs

    ds = zds.ZarrDataset(
        dataset_specs=dataset_specs,
        shuffle=shuffle,
        patch_sampler=patch_sampler,
        draw_same_chunk=draw_same_chunk
    )

    array_idx = 0
    label_idx = 1

    n_samples = 0

    for sample in ds:
        n_samples += 1
        assert isinstance(sample, tuple), \
            (f"When `return_positions`, `return_worker_id` or a labels dataset"
             f" specification is passed to ZarrDataset, retrieved samples "
             f"should be a tuple, got {type(sample)} instead.")

        assert isinstance(sample[0], np.ndarray), \
            (f"Sample should be a Numpy NDArray, got {type(sample[0])} "
             f"instead.")

        sample_array = sample[array_idx]
        labels_array = sample[label_idx]

        expected_shape = tuple(
            min(patch_size, specs["shape"][specs["source_axes"].index(ax)])
            if ax in patch_sampler.spatial_axes else
            specs["shape"][specs["source_axes"].index(ax)]
            for ax in specs["source_axes"]
        )

        expected_labels_shape = tuple(
            patch_size if ax in patch_sampler.spatial_axes else
            specs["shape"]["YX".index(ax)]
            for ax in "YX"
        )

        assert tuple(sample_array.shape) == expected_shape, \
            (f"Sample expected to have shape {expected_shape}, got "
             f"{sample_array.shape} instead")

        assert tuple(labels_array.shape) == expected_labels_shape, \
            (f"Labels expected to have shape {expected_labels_shape}, got "
             f"{labels_array.shape} instead")

    assert n_samples > 0, ("Expected more than zero samples extracted from "
                           "this experiment.")

    # Second iteration
    n_samples = 0
    for sample in ds:
        n_samples += 1
        assert isinstance(sample, tuple), \
            (f"When `return_positions`, `return_worker_id` or a labels dataset"
             f" specification is passed to ZarrDataset, retrieved samples "
             f"should be a tuple, got {type(sample)} instead.")

        assert isinstance(sample[0], np.ndarray), \
            (f"Sample should be a Numpy NDArray, got {type(sample[0])} "
             f"instead.")

        sample_array = sample[array_idx]
        labels_array = sample[label_idx]

        expected_shape = tuple(
            min(patch_size, specs["shape"][specs["source_axes"].index(ax)])
            if ax in patch_sampler.spatial_axes else
            specs["shape"][specs["source_axes"].index(ax)]
            for ax in specs["source_axes"]
        )

        expected_labels_shape = tuple(
            patch_size if ax in patch_sampler.spatial_axes else
            specs["shape"]["YX".index(ax)]
            for ax in "YX"
        )

        assert tuple(sample_array.shape) == expected_shape, \
            (f"Sample expected to have shape {expected_shape}, got "
             f"{sample_array.shape} instead")

        assert tuple(labels_array.shape) == expected_labels_shape, \
            (f"Labels expected to have shape {expected_labels_shape}, got "
             f"{labels_array.shape} instead")

    assert n_samples > 0, ("Expected more than zero samples extracted from "
                           "this experiment.")


@pytest.mark.parametrize(
    "image_dataset_specs, patch_sampler_specs", [
        (IMAGE_SPECS[10], (1024, True)),
        (IMAGE_SPECS[10], (1024, False)),
    ],
    indirect=["image_dataset_specs", "patch_sampler_specs"]
)
def test_greater_patch_ZarrDataset(image_dataset_specs, patch_sampler_specs):
    dataset_specs, specs = image_dataset_specs
    patch_sampler, patch_size, allow_incomplete_patches = patch_sampler_specs

    ds = zds.ZarrDataset(
        dataset_specs=dataset_specs,
        patch_sampler=patch_sampler
    )

    n_samples = 0
    for _ in ds:
        n_samples += 1

    if allow_incomplete_patches:
        assert n_samples > 0, ("Expected at elast one sample when patch"
                               " size is greater than the image size, and"
                               " `allow_incomplete_patches` is True.")
    else:
        assert n_samples == 0, ("Expected zero samples since requested patch"
                                " size is greater than the image size, and"
                                " `allow_incomplete_patches` is False.")


@pytest.mark.parametrize(
    "image_dataset_specs, patch_sampler_specs, shuffle, draw_same_chunk,"
    "batch_size, num_workers", [
        (IMAGE_SPECS[10], (32, False), True, False, 2, 2),
        ([IMAGE_SPECS[10]] * 4, (32, False), True, True, 2, 3),
        ([IMAGE_SPECS[10]] * 2, (32, False), True, True, 2, 3),
    ],
    indirect=["image_dataset_specs", "patch_sampler_specs"]
)
def test_multithread_ZarrDataset(image_dataset_specs, patch_sampler_specs,
                                 shuffle,
                                 draw_same_chunk,
                                 batch_size,
                                 num_workers):
    dataset_specs, specs = image_dataset_specs

    patch_sampler, patch_size, allow_incomplete_patches = patch_sampler_specs

    ds = zds.ZarrDataset(
        dataset_specs=dataset_specs,
        shuffle=shuffle,
        patch_sampler=patch_sampler,
        draw_same_chunk=draw_same_chunk
    )

    dl = DataLoader(
        ds,
        batch_size=batch_size,
        num_workers=num_workers,
        worker_init_fn=zds.zarrdataset_worker_init_fn,
        drop_last=True
    )

    array_idx = 0
    label_idx = 1

    n_samples = 0

    for sample in dl:
        n_samples += 1
        assert isinstance(sample, list), \
            (f"When a labels dataset specification is passed to ZarrDataset "
             f"and used with PyTorch DataLoader, samples should be a list, got"
             f" {type(sample)} instead.")

        assert isinstance(sample[0], torch.Tensor), \
            (f"Sample should be a PyTorch Tensor, got {type(sample[0])} "
             f"instead.")

        sample_array = sample[array_idx]
        labels_array = sample[label_idx]

        expected_shape = [batch_size] + [
            min(patch_size, specs["shape"][specs["source_axes"].index(ax)])
            if ax in patch_sampler.spatial_axes else
            specs["shape"][specs["source_axes"].index(ax)]
            for ax in specs["source_axes"]
        ]

        expected_labels_shape = [batch_size] + [
            patch_size if ax in patch_sampler.spatial_axes else
            specs["shape"]["YX".index(ax)]
            for ax in "YX"
        ]

        assert all(map(operator.eq,
                       sample_array.shape,
                       expected_shape)), \
            (f"Sample expected to have shape {expected_shape}, got "
             f"{sample_array.shape} instead")

        assert all(map(operator.eq,
                       labels_array.shape,
                       expected_labels_shape)), \
            (f"Labels expected to have shape {expected_labels_shape}, got "
             f"{labels_array.shape} instead")

    assert n_samples > 0, ("Expected more than zero samples extracted from "
                           "this experiment.")


@pytest.mark.parametrize(
    "image_dataset_specs, patch_sampler_specs, shuffle, draw_same_chunk,"
    "batch_size, num_workers, repeat_dataset", [
        (IMAGE_SPECS[10:12], (32, False), True, False, 2, 2, 1),
        (IMAGE_SPECS[10:12], (32, False), True, False, 2, 2, 2),
        (IMAGE_SPECS[10:12], (32, False), True, False, 2, 2, 3),
    ],
    indirect=["image_dataset_specs", "patch_sampler_specs"]
)
def test_multithread_chained_ZarrDataset(image_dataset_specs,
                                         patch_sampler_specs,
                                         shuffle,
                                         draw_same_chunk,
                                         batch_size,
                                         num_workers,
                                         repeat_dataset):
    dataset_specs, specs = image_dataset_specs
    patch_sampler, patch_size, allow_incomplete_patches = patch_sampler_specs

    ds = [zds.ZarrDataset(dataset_specs=dataset_specs,
                          shuffle=shuffle,
                          patch_sampler=patch_sampler,
                          draw_same_chunk=draw_same_chunk
        )] * repeat_dataset

    chained_ds = ChainDataset(ds)

    dl = DataLoader(
        chained_ds,
        batch_size=batch_size,
        num_workers=num_workers,
        worker_init_fn=zds.chained_zarrdataset_worker_init_fn,
        drop_last=True
    )

    array_idx = 0
    label_idx = 1

    n_samples = 0

    for sample in dl:
        n_samples += 1
        assert isinstance(sample, list), \
            (f"When a labels dataset specification is passed to ZarrDataset "
             f"and used with PyTorch DataLoader, samples should be a list, got"
             f" {type(sample)} instead.")

        assert isinstance(sample[0], torch.Tensor), \
            (f"Sample should be a PyTorch Tensor, got {type(sample[0])} "
             f"instead.")

        sample_array = sample[array_idx]
        labels_array = sample[label_idx]

        expected_shape = [batch_size] + [
            min(patch_size, specs["shape"][specs["source_axes"].index(ax)])
            if ax in patch_sampler.spatial_axes else
            specs["shape"][specs["source_axes"].index(ax)]
            for ax in specs["source_axes"]
        ]

        expected_labels_shape = [batch_size] + [
            patch_size if ax in patch_sampler.spatial_axes else
            specs["shape"]["YX".index(ax)]
            for ax in "YX"
        ]

        assert all(map(operator.eq,
                       list(sample_array.shape),
                       expected_shape)), \
            (f"Sample expected to have shape {expected_shape}, got "
             f"{sample_array.shape} instead")

        assert all(map(operator.eq,
                       labels_array.shape,
                       expected_labels_shape)), \
            (f"Labels expected to have shape {expected_labels_shape}, got "
             f"{labels_array.shape} instead")

    assert n_samples > 0, ("Expected more than zero samples extracted from "
                           "this experiment.")


def test_compatibility_no_pytroch():
    with mock.patch.dict('sys.modules', {'torch': None}):
        importlib.reload(zds._zarrdataset)

        dataset = zds._zarrdataset.ZarrDataset()

        assert isinstance(object, type(dataset).__bases__), \
            (f"When pytorch is not installed, ZarrDataset should be inherited"
             f" from object, not {type(dataset).__bases__}")

        try:
            zds._zarrdataset.zarrdataset_worker_init_fn(None)

        except Exception as e:
            raise AssertionError(f"No exceptions where expected when using "
                                 f"`zarrdataset_worker_init_fn` without "
                                 f"pytorch installed, got {e} instead.")

        try:
            zds._zarrdataset.chained_zarrdataset_worker_init_fn(None)

        except Exception as e:
            raise AssertionError(f"No exceptions where expected when using "
                                 f"`chained_zarrdataset_worker_init_fn` "
                                 f"without pytorch installed, got {e} "
                                 f"instead.")
    with mock.patch.dict('sys.modules', {'torch': torch}):
        importlib.reload(zds._zarrdataset)
        importlib.reload(zds)



import zarrdataset as zds
import numpy as np


def test_ToDtype():
    transform = zds.ToDtype(dtype=np.float32)

    a = np.empty((0, ), dtype=np.int16)

    b = transform(a)

    assert b.dtype == np.float32, \
        (f"Expected array after transformation be of type np.float32, got "
         f"{b.dtype} instead.")

    assert "dtype=<class 'numpy.float32'>" in str(transform), \
        (f"Expected string generated from the transform to include "
         f"'dtype=np.float32', got {str(transform)} instead.")



import pytest
import os
import shutil
from pathlib import Path
import operator

from skimage import transform
from tests.utils import IMAGE_SPECS, MASKABLE_IMAGE_SPECS

import zarrdataset as zds
import math
import numpy as np


@pytest.fixture(scope="function")
def image_collection(request):
    dst_dir = request.param["dst_dir"]

    if dst_dir is not None:
        dst_dir = Path(request.param["dst_dir"])
        dst_dir.mkdir(parents=True, exist_ok=True)

    mask_args = None
    if not isinstance(request.param["source"], str):
        (img_src,
         mask_src,
         labels_src,
         classes_src) = request.param["source"](request.param["dst_dir"],
                                                request.param["specs"])

        if mask_src is not None:
            mask_args = dict(
                filename=mask_src,
                source_axes="YX",
                data_group=request.param["specs"]["mask_group"],
            )

    else:
        img_src = request.param["source"]

    collection_args = dict(
        images=dict(
            filename=img_src,
            source_axes=request.param["specs"]["source_axes"],
            axes=request.param["specs"].get("axes", None),
            data_group=request.param["specs"].get("data_group", None),
            roi=request.param["specs"].get("roi", None),
        ),
    )

    if mask_args is not None:
        collection_args["masks"] = mask_args

    image_collection = zds.ImageCollection(collection_args=collection_args)

    yield image_collection

    if dst_dir is not None and os.path.isdir(dst_dir):
        shutil.rmtree(dst_dir)


@pytest.fixture(scope="function")
def image_collection_mask_not2scale(request):
    dst_dir = request.param["dst_dir"]

    if dst_dir is not None:
        dst_dir = Path(request.param["dst_dir"])
        dst_dir.mkdir(parents=True, exist_ok=True)

    (img_src,
     mask_src,
     labels_src,
     classes_src) = request.param["source"](request.param["dst_dir"],
                                            request.param["specs"])

    collection_args = dict(
        images=dict(
            filename=img_src,
            source_axes=request.param["specs"]["source_axes"],
            data_group=request.param["specs"]["data_group"],
        ),
        masks=dict(
            filename=np.ones((3, 5), dtype=bool),
            source_axes="YX",
            data_group=None,
        )
    )

    image_collection = zds.ImageCollection(collection_args=collection_args)

    yield image_collection

    if dst_dir is not None and os.path.isdir(dst_dir):
        shutil.rmtree(dst_dir)


@pytest.mark.parametrize("patch_size", [
    512,
])
def test_string_PatchSampler(patch_size):
    patch_sampler = zds.PatchSampler(patch_size=patch_size)
    ps_str_repr = str(patch_sampler)

    assert f"patches of size {patch_sampler._patch_size}" in ps_str_repr,\
        (f"Expected string representation of PatchSampler to show the patch"
         f" size, got `{ps_str_repr}` instead.")

    patch_sampler = zds.BlueNoisePatchSampler(patch_size=patch_size)
    ps_str_repr = str(patch_sampler)

    assert f"patches of size {patch_sampler._patch_size}" in ps_str_repr,\
        (f"Expected string representation of BlueNoisePatchSampler to show the"
         f" patch size, got `{ps_str_repr}` instead.")


@pytest.mark.parametrize("patch_size, spatial_axes, expected_patch_size", [
    (512, "X", dict(X=512)),
    ((128, 64), "XY", dict(X=128, Y=64)),
])
def test_PatchSampler_correct_patch_size(patch_size, spatial_axes,
                                         expected_patch_size):
    patch_sampler = zds.PatchSampler(patch_size=patch_size,
                                     spatial_axes=spatial_axes)

    assert patch_sampler._patch_size == expected_patch_size, \
        (f"Expected `patch_size` to be a dictionary as {expected_patch_size}, "
         f"got {patch_sampler._patch_size} instead.")


@pytest.mark.parametrize("stride, spatial_axes, expected_stride", [
    (512, "X", dict(X=512)),
    ((128, 64), "XY", dict(X=128, Y=64)),
])
def test_PatchSampler_correct_stride(stride, spatial_axes, expected_stride):
    patch_sampler = zds.PatchSampler(patch_size=512, stride=stride,
                                     spatial_axes=spatial_axes)

    assert patch_sampler._stride == expected_stride, \
        (f"Expected `stride` to be a dictionary as {expected_stride}, "
         f"got {patch_sampler._stride} instead.")


@pytest.mark.parametrize("pad, spatial_axes, expected_pad", [
    (512, "X", dict(X=512)),
    ((128, 64), "XY", dict(X=128, Y=64)),
])
def test_PatchSampler_correct_pad(pad, spatial_axes, expected_pad):
    patch_sampler = zds.PatchSampler(patch_size=512, pad=pad,
                                     spatial_axes=spatial_axes)

    assert patch_sampler._pad == expected_pad, \
        (f"Expected `pad` to be a dictionary as {expected_pad}, "
         f"got {patch_sampler._pad} instead.")


@pytest.mark.parametrize("patch_size, spatial_axes", [
    ((512, 128), "X"),
    ((128, ), "XY"),
    ("patch_size", "ZYX"),
])
def test_PatchSampler_incorrect_patch_size(patch_size, spatial_axes):
    with pytest.raises(ValueError):
        patch_sampler = zds.PatchSampler(patch_size=patch_size,
                                         spatial_axes=spatial_axes)


@pytest.mark.parametrize("stride, spatial_axes", [
    ((512, 128), "X"),
    ((128, ), "XY"),
    ("stride", "ZYX"),
])
def test_PatchSampler_incorrect_stride(stride, spatial_axes):
    with pytest.raises(ValueError):
        patch_sampler = zds.PatchSampler(patch_size=512,
                                         stride=stride,
                                         spatial_axes=spatial_axes)


@pytest.mark.parametrize("pad, spatial_axes", [
    ((512, 128), "X"),
    ((128, ), "XY"),
    ("pad", "ZYX"),
])
def test_PatchSampler_incorrect_pad(pad, spatial_axes):
    with pytest.raises(ValueError):
        patch_sampler = zds.PatchSampler(patch_size=512,
                                         pad=pad,
                                         spatial_axes=spatial_axes)


@pytest.mark.parametrize("patch_size, image_collection", [
    (32, IMAGE_SPECS[10])
], indirect=["image_collection"])
def test_PatchSampler_chunk_generation(patch_size, image_collection):
    patch_sampler = zds.PatchSampler(patch_size)

    chunks_toplefts = patch_sampler.compute_chunks(image_collection)

    chunk_size = {
        ax: cs
        for ax, cs in zip(image_collection.collection["images"].axes,
                          image_collection.collection["images"].chunk_size)
    }

    scaled_chunk_size = {
        ax: int(cs * image_collection.collection["masks"].scale[ax])
        for ax, cs in zip(image_collection.collection["images"].axes,
                          image_collection.collection["images"].chunk_size)
        if ax in image_collection.collection["masks"].axes
    }

    scaled_mask = transform.downscale_local_mean(
        image_collection.collection["masks"][:],
        factors=(scaled_chunk_size["Y"], scaled_chunk_size["X"])
    )
    expected_chunks_toplefts = np.nonzero(scaled_mask)

    expected_chunks_toplefts = [
        dict(
            [("Z", slice(0, 1, None))]
            + [
               (ax, slice(tl * chunk_size[ax], (tl + 1) * chunk_size[ax]))
               for ax, tl in zip("YX", tls)
            ]
        )
        for tls in zip(*expected_chunks_toplefts)
    ]

    assert all(map(operator.eq, chunks_toplefts, expected_chunks_toplefts)), \
        (f"Expected chunks to be {expected_chunks_toplefts[:3]}, got "
         f"{chunks_toplefts[:3]} instead.")


@pytest.mark.parametrize("patch_size, image_collection", [
    (32, IMAGE_SPECS[10])
], indirect=["image_collection"])
def test_PatchSampler(patch_size, image_collection):
    patch_sampler = zds.PatchSampler(patch_size)

    chunks_toplefts = patch_sampler.compute_chunks(image_collection)

    patches_toplefts = patch_sampler.compute_patches(
        image_collection,
        chunk_tlbr=chunks_toplefts[0]
    )

    scaled_patch_size = {
        ax: int(patch_size * scl)
        for ax, scl in image_collection.collection["masks"].scale.items()
    }

    scaled_mask = transform.downscale_local_mean(
        image_collection.collection["masks"][chunks_toplefts[0]],
        factors=(scaled_patch_size["Y"], scaled_patch_size["X"])
    )
    expected_patches_toplefts = np.nonzero(scaled_mask)

    expected_patches_toplefts = [
        dict(
            [("Z", slice(0, 1, None))]
            + [
               (ax, slice(tl * patch_size, (tl + 1) * patch_size))
               for ax, tl in zip("YX", tls)
            ]
        )
        for tls in zip(*expected_patches_toplefts)
    ]

    assert all(map(operator.eq, patches_toplefts, expected_patches_toplefts)),\
        (f"Expected patches to be {expected_patches_toplefts[:3]}, got "
         f"{patches_toplefts[:3]} instead.")


@pytest.mark.parametrize("patch_size, stride, image_collection", [
    (32, 32, IMAGE_SPECS[10]),
    (32, 16, IMAGE_SPECS[10]),
    (32, 64, IMAGE_SPECS[10]),
], indirect=["image_collection"])
def test_PatchSampler_stride(patch_size, stride, image_collection):
    patch_sampler = zds.PatchSampler(patch_size, stride=stride)

    chunks_toplefts = patch_sampler.compute_chunks(image_collection)

    patches_toplefts = patch_sampler.compute_patches(
        image_collection,
        chunk_tlbr=chunks_toplefts[0]
    )

    scaled_patch_size = {
        ax: int(stride * scl)
        for ax, scl in image_collection.collection["masks"].scale.items()
    }

    scaled_mask = transform.downscale_local_mean(
        image_collection.collection["masks"][chunks_toplefts[0]],
        factors=(scaled_patch_size["Y"], scaled_patch_size["X"])
    )
    expected_patches_toplefts = np.nonzero(scaled_mask)

    expected_patches_toplefts = [
        dict(
            [("Z", slice(0, 1, None))]
            + [
               (ax, slice(tl * stride, tl * stride + patch_size))
               for ax, tl in zip("YX", tls)
            ]
        )
        for tls in zip(*expected_patches_toplefts)
    ]
    assert all(map(operator.eq, patches_toplefts, expected_patches_toplefts)),\
        (f"Expected patches to be {expected_patches_toplefts[:3]}, got "
         f"{patches_toplefts[:3]} instead.")


@pytest.mark.parametrize("patch_size, pad, image_collection", [
    (32, 0, IMAGE_SPECS[10]),
    (32, 2, IMAGE_SPECS[10]),
], indirect=["image_collection"])
def test_PatchSampler_pad(patch_size, pad, image_collection):
    patch_sampler = zds.PatchSampler(patch_size, pad=pad)

    chunks_toplefts = patch_sampler.compute_chunks(image_collection)

    patches_toplefts = patch_sampler.compute_patches(
        image_collection,
        chunk_tlbr=chunks_toplefts[0]
    )

    scaled_patch_size = {
        ax: int(patch_size * scl)
        for ax, scl in image_collection.collection["masks"].scale.items()
    }

    scaled_mask = transform.downscale_local_mean(
        image_collection.collection["masks"][chunks_toplefts[0]],
        factors=(scaled_patch_size["Y"], scaled_patch_size["X"])
    )
    expected_patches_toplefts = np.nonzero(scaled_mask)

    # TODO: Change expected patches toplefts for strided ones
    expected_patches_toplefts = [
        dict(
            [("Z", slice(0, 1, None))]
            + [
               (ax, slice(tl * patch_size - pad, (tl + 1) * patch_size + pad))
               for ax, tl in zip("YX", tls)
            ]
        )
        for tls in zip(*expected_patches_toplefts)
    ]

    assert all(map(operator.eq, patches_toplefts, expected_patches_toplefts)),\
        (f"Expected patches to be {expected_patches_toplefts[:3]}, got "
         f"{patches_toplefts[:3]} instead.")


@pytest.mark.parametrize("patch_size, allow_incomplete_patches,"
                         "image_collection", [
    (1024, True, IMAGE_SPECS[10]),
    (1024, False, IMAGE_SPECS[10]),
], indirect=["image_collection"])
def test_PatchSampler_incomplete_patches(patch_size, allow_incomplete_patches,
                                         image_collection):
    patch_sampler = zds.PatchSampler(
        patch_size,
        allow_incomplete_patches=allow_incomplete_patches
    )

    chunks_toplefts = patch_sampler.compute_chunks(image_collection)

    patches_toplefts = patch_sampler.compute_patches(
        image_collection,
        chunk_tlbr=chunks_toplefts[0]
    )

    expected_num_patches = 1 if allow_incomplete_patches else 0

    assert len(patches_toplefts) == expected_num_patches,\
        (f"Expected to have {expected_num_patches}, when "
         f"`allow_incomplete_patches` is {allow_incomplete_patches} "
         f"got {len(patches_toplefts)} instead.")


@pytest.mark.parametrize("patch_size, axes, resample, allow_overlap,"
                         "image_collection", [
    (dict(X=32, Y=32, Z=1), "XYZ", True, True, IMAGE_SPECS[10]),
    (dict(X=32, Y=32), "YX", False, False, IMAGE_SPECS[10]),
], indirect=["image_collection"])
def test_BlueNoisePatchSampler(patch_size, axes, resample, allow_overlap,
                               image_collection):
    np.random.seed(447788)

    patch_sampler = zds.BlueNoisePatchSampler(patch_size,
                                              resample_positions=resample,
                                              allow_overlap=allow_overlap,
                                              spatial_axes=axes)

    chunks_toplefts = patch_sampler.compute_chunks(image_collection)

    patches_toplefts = patch_sampler.compute_patches(
        image_collection,
        chunk_tlbr=chunks_toplefts[0]
    )

    assert len(patches_toplefts) == len(patch_sampler._base_chunk_tls), \
        (f"Expected {len(patch_sampler._base_chunk_tls)} patches, got "
         f"{len(patches_toplefts)} instead.")


@pytest.mark.parametrize("image_collection_mask_not2scale", [
    IMAGE_SPECS[10]
], indirect=["image_collection_mask_not2scale"])
def test_BlueNoisePatchSampler_mask_not2scale(image_collection_mask_not2scale):
    np.random.seed(447788)

    patch_size = dict(X=1024, Y=1024)

    patch_sampler = zds.BlueNoisePatchSampler(patch_size)

    chunks_toplefts = patch_sampler.compute_chunks(
        image_collection_mask_not2scale
    )

    patches_toplefts = patch_sampler.compute_patches(
        image_collection_mask_not2scale,
        chunk_tlbr=chunks_toplefts[0]
    )

    # Samples can be retrieved from chunks that are not multiple of the patch
    # size. The ZarrDataset class should handle these cases, either by droping
    # these patches, or by adding padding when allowed by the user.
    assert len(patches_toplefts) == 1, \
        (f"Expected 0 patches, got {len(patches_toplefts)} instead.")


@pytest.mark.parametrize("patch_size, stride, image_collection, specs", [
    (512, 512, MASKABLE_IMAGE_SPECS[0], MASKABLE_IMAGE_SPECS[0]),
    (512, 256, MASKABLE_IMAGE_SPECS[0], MASKABLE_IMAGE_SPECS[0])
], indirect=["image_collection"])
def test_unique_sampling_PatchSampler(patch_size, stride, image_collection,
                                      specs):
    from skimage import color, filters, morphology
    import zarr

    z_img = zarr.open(specs["source"], mode="r")

    im_gray = color.rgb2gray(z_img["4"][0, :, 0], channel_axis=0)
    thresh = filters.threshold_otsu(im_gray)

    mask = im_gray > thresh
    mask = morphology.remove_small_objects(mask == 0, min_size=16 ** 2,
                                           connectivity=2)
    mask = morphology.remove_small_holes(mask, area_threshold=128)
    mask = morphology.binary_erosion(mask, morphology.disk(8))
    mask = morphology.binary_dilation(mask, morphology.disk(8))

    image_collection.collection["masks"] = zds.ImageLoader(filename=mask,
                                                           source_axes="YX",
                                                           mode="masks")
    image_collection.reset_scales()

    patch_sampler = zds.PatchSampler(patch_size, stride=stride,
                                     min_area=1/16 ** 2)

    chunks_toplefts = patch_sampler.compute_chunks(image_collection)

    all_patches_tls = []
    all_chunks_tls = []

    for ctl in chunks_toplefts:
        patches_toplefts = patch_sampler.compute_patches(image_collection,
                                                         chunk_tlbr=ctl)

        for ptl in patches_toplefts:
            assert ptl not in all_patches_tls,\
                (f"Expected no repetitions in patch sampling, got {ptl} "
                 f"twice instead at chunk {ctl}. Possible duplicate of patch "
                 f"{all_patches_tls.index(ptl)}, from chunk "
                 f"{all_chunks_tls[all_patches_tls.index(ptl)]}")
            all_patches_tls.append(ptl)
            all_chunks_tls.append(ctl)


@pytest.mark.parametrize("patch_size, image_collection, specs", [
    (512, MASKABLE_IMAGE_SPECS[0], MASKABLE_IMAGE_SPECS[0])
], indirect=["image_collection"])
def test_unique_sampling_BlueNoisePatchSampler(patch_size, image_collection,
                                               specs):
    from skimage import color, filters, morphology
    import zarr

    z_img = zarr.open(specs["source"], mode="r")

    im_gray = color.rgb2gray(z_img["4"][0, :, 0], channel_axis=0)
    thresh = filters.threshold_otsu(im_gray)

    mask = im_gray > thresh
    mask = morphology.remove_small_objects(mask == 0, min_size=16 ** 2,
                                           connectivity=2)
    mask = morphology.remove_small_holes(mask, area_threshold=128)
    mask = morphology.binary_erosion(mask, morphology.disk(8))
    mask = morphology.binary_dilation(mask, morphology.disk(8))

    image_collection.collection["masks"] = zds.ImageLoader(filename=mask,
                                                           source_axes="YX",
                                                           mode="masks")
    image_collection.reset_scales()

    patch_sampler = zds.BlueNoisePatchSampler(patch_size)

    chunks_toplefts = patch_sampler.compute_chunks(image_collection)

    all_patches_tls = []
    all_chunks_tls = []

    for ctl in chunks_toplefts:
        patches_toplefts = patch_sampler.compute_patches(image_collection,
                                                         chunk_tlbr=ctl)

        for ptl in patches_toplefts:
            assert ptl not in all_patches_tls,\
                (f"Expected no repetitions in patch sampling, got {ptl} "
                 f"twice instead at chunk {ctl}. Possible duplicate of patch "
                 f"{all_patches_tls.index(ptl)}, from chunk "
                 f"{all_chunks_tls[all_patches_tls.index(ptl)]}")
            all_patches_tls.append(ptl)
            all_chunks_tls.append(ctl)


@pytest.mark.parametrize("min_area, patch_size, "
                         "image_collection_mask_not2scale", [
    (0.5, 32, IMAGE_SPECS[11]),
    (4096, 512, IMAGE_SPECS[11]),
    (0.5, 1024, IMAGE_SPECS[11])
], indirect=["image_collection_mask_not2scale"])
def test_min_area_sampling_PatchSampler(min_area, patch_size,
                                        image_collection_mask_not2scale):
    patch_sampler = zds.PatchSampler(patch_size, min_area=min_area)

    chunks_toplefts = patch_sampler.compute_chunks(
        image_collection_mask_not2scale
    )

    all_patches_tls = []
    all_chunks_tls = []

    for ctl in chunks_toplefts:
        patches_toplefts = patch_sampler.compute_patches(
            image_collection_mask_not2scale,
            chunk_tlbr=ctl
        )

        for ptl in patches_toplefts:
            assert ptl not in all_patches_tls,\
                (f"Expected no repetitions in patch sampling, got {ptl} "
                 f"twice instead at chunk {ctl}. Possible duplicate of patch "
                 f"{all_patches_tls.index(ptl)}, from chunk "
                 f"{all_chunks_tls[all_patches_tls.index(ptl)]}")
            all_patches_tls.append(ptl)
            all_chunks_tls.append(ctl)
